import os
import shutil
from typing import Optional, List

from pypaq.exception import PyPaqException
from pypaq.lipytools.printout import stamp
from pypaq.lipytools.files import r_pickle, w_pickle, prep_folder
from pypaq.lipytools.pylogger import get_pylogger
from pypaq.pms.base import POINT
from pypaq.pms.para import ParaGX


class ParaSaveException(PyPaqException):
    pass


class ParaSave(ParaGX):
    """ Parameters Save
    saves/loads POINT (self params) from save_topdir
    implements GX for saved

    folders functionality (load, save ..) is disabled for save_topdir / SAVE_TOPDIR == None

    ParaSave defaults are stored in PARASAVE_DEFAULTS dict and cannot be placed in __init__ defaults.
    This is a consequence of the params resolution mechanism in ParaSave,
    where parameters may come from three sources, and each subsequent source overrides the previous ones:
        1. class __init__ defaults
        2. saved in the folder
        3. given with kwargs in __init__
    If all ParaSave parameters were set with __init__ defaults,
    it would not be possible to distinguish between sources 1 and 3.

    @DynamicAttrs <-- disables warning for unresolved attributes references """

    PARASAVE_DEFAULTS = {
        'gxable':   True,
        'parents':  [],     # list (nested) of parents names
    }

    SAVE_TOPDIR = None      # save top directory
    SAVE_FN_PFX = 'point'   # POINT file prefix
    OBJ_SUFFIX = '.dct'     # POINT file suffix of obj(pickle)
    TXT_SUFFIX = '.txt'     # POINT file suffix of text

    def __init__(
            self,
            name: str,
            save_topdir: Optional[str]= None,   # ParaSave top directory
            save_fn_pfx: Optional[str]= None,   # ParaSave POINT file prefix
            assert_saved=               False,  # for True asserts that ParaSave has been already saved in save_topdir
            lock_managed_params=        False,  # locks _managed_params to only those known while init
            logger=                     None,
            loglevel=                   20,
            **kwargs,
    ):
        """ name, save_topdir, save_fn_pfx -> if given -> always override saved values """

        # _managed_params allows to lock ParaSave managed params only to those resolved here (while __init__)
        self._managed_params: Optional[List[str]] = None

        self.name = name
        self.save_topdir = save_topdir or self.SAVE_TOPDIR
        self.save_fn_pfx = save_fn_pfx or self.SAVE_FN_PFX

        if not logger:
            logger = get_pylogger(
                name=       self.name,
                add_stamp=  False,
                folder=     ParaSave._full_dir(name=self.name, save_topdir=self.save_topdir) if self.save_topdir else None,
                level=      loglevel)
        self.__log = logger

        self.__log.info(f'*** ParaSave : {self.name} *** initializes..')
        self.__log.debug(f'> save_topdir: {self.save_topdir}')
        self.__log.debug(f'> save_fn_pfx: {self.save_fn_pfx}')

        if assert_saved and not self.is_saved(
                name=           self.name,
                save_topdir=    self.save_topdir,
                save_fn_pfx=    self.save_fn_pfx):
            ex_msg = f'ParaSave {self.name} does not exist, but should!'
            self.__log.error(ex_msg)
            raise ParaSaveException(ex_msg)

        point_saved = self.load_point(
            name=           self.name,
            save_topdir=    self.save_topdir,
            save_fn_pfx=    self.save_fn_pfx)

        # update in proper order
        _self_point = {}
        _self_point.update(self.PARASAVE_DEFAULTS)
        _self_point.update(point_saved)
        _self_point.update(self.get_point()) # params added up to now to self
        _self_point.update(kwargs)

        self.__log.debug(f'ParaSave POINT sources:')
        self.__log.debug(f'> PARASAVE_DEFAULTS:     {self.PARASAVE_DEFAULTS}')
        self.__log.debug(f'> POINT saved:           {point_saved}')
        self.__log.debug(f'> given kwargs:          {kwargs}')
        self.__log.debug(f'ParaSave complete POINT: {_self_point}')

        super().__init__(**_self_point)

        if lock_managed_params:
            self._managed_params = self.get_managed_params()
            self.__log.debug(f'locked managed params: {self._managed_params}')

    def get_managed_params(self) -> List[str]:
        if self._managed_params is not None:
            return self._managed_params
        return ParaGX.get_managed_params(self)

    def __setitem__(self, key, value):
        if self._managed_params is not None:
            msg = f'ParaSave (managed_params locked) asked to self set with: >{key}:{value}< self will be updated, but managed_params not!'
            self.__log.warning(msg)
        setattr(self, key, value)

    # ************************************************************************************************ folder management

    @classmethod
    def _full_dir(
            cls,
            name: str,
            save_topdir: Optional[str]= None,
    ):
        if not save_topdir: save_topdir = cls.SAVE_TOPDIR
        return f'{save_topdir}/{name}'

    @classmethod
    def _obj_fn(
            cls,
            name: str,
            save_topdir: Optional[str]= None,
            save_fn_pfx: Optional[str]= None,
    ):
        if not save_topdir: save_topdir = cls.SAVE_TOPDIR
        if not save_fn_pfx: save_fn_pfx = cls.SAVE_FN_PFX
        return f'{cls._full_dir(name=name, save_topdir=save_topdir)}/{save_fn_pfx}{cls.OBJ_SUFFIX}'

    @classmethod
    def _txt_fn(
            cls,
            name: str,
            save_topdir: Optional[str]= None,
            save_fn_pfx: Optional[str]= None,
    ):
        if not save_topdir: save_topdir = cls.SAVE_TOPDIR
        if not save_fn_pfx: save_fn_pfx = cls.SAVE_FN_PFX
        return f'{cls._full_dir(name=name, save_topdir=save_topdir)}/{save_fn_pfx}{cls.TXT_SUFFIX}'

    @classmethod
    def is_saved(
            cls,
            name: str,
            save_topdir: Optional[str]= None,
            save_fn_pfx: Optional[str]= None,
    ) -> bool:
        """ checks if given ParaSave name is already saved (..has been created before) """
        if not save_topdir: save_topdir = cls.SAVE_TOPDIR
        if not save_fn_pfx: save_fn_pfx = cls.SAVE_FN_PFX
        return os.path.isfile(cls._obj_fn(name=name, save_topdir=save_topdir, save_fn_pfx=save_fn_pfx))

    @classmethod
    def load_point(
            cls,
            name: str,
            save_topdir: Optional[str]= None,
            save_fn_pfx: Optional[str]= None,
    ) -> POINT:
        """ loads POINT from folder """
        obj_FN = cls._obj_fn(
            name=           name,
            save_topdir=    save_topdir,
            save_fn_pfx=    save_fn_pfx)
        if os.path.isfile(obj_FN):
            point: POINT = r_pickle(obj_FN)
            return point
        return {}

    def save_point(self):
        """ saves self POINT to folder (with preview in txt) """

        if not self.save_topdir:
            msg = f'cannot save {self.__class__.__name__}, if save directory was not given, aborting'
            self.__log.error(msg)
            raise ParaSaveException(msg)

        obj_FN = self._obj_fn(
            name=           self.name,
            save_topdir=    self.save_topdir,
            save_fn_pfx=    self.save_fn_pfx)
        txt_FN = self._txt_fn(
            name=           self.name,
            save_topdir=    self.save_topdir,
            save_fn_pfx=    self.save_fn_pfx)

        prep_folder(self._full_dir(
            name=           self.name,
            save_topdir=    self.save_topdir))

        if os.path.isfile(obj_FN): shutil.copy(obj_FN, f'{obj_FN}_OLD')
        if os.path.isfile(txt_FN): shutil.copy(txt_FN, f'{txt_FN}_OLD')

        point: POINT = self.get_point()
        w_pickle(point, obj_FN)
        with open(txt_FN, 'w') as file:
            s = f' *** ParaSave POINT saved: {stamp(year=True, letters=None)} ***\n'
            s += self.dict_2str(point)
            file.write(s)

        self.__log.debug(f'{self.__class__.__name__} {self.name} saved to {self.save_topdir}')

    @classmethod
    def oversave_point(
            cls,
            name: str,
            save_topdir: Optional[str]= None,
            save_fn_pfx: Optional[str]= None,
            **kwargs,
    ):
        """ loads, next overrides parameters from given kwargs and saves POINT """
        psc = cls(
            name=           name,
            save_topdir=    save_topdir,
            save_fn_pfx=    save_fn_pfx,
            loglevel=       30)
        psc.update(kwargs)
        psc.save_point()

    @classmethod
    def copy_saved_point(
            cls,
            name_src: str,
            name_trg: str,
            save_topdir_src: Optional[str]= None,
            save_topdir_trg: Optional[str]= None,
            save_fn_pfx: Optional[str]=     None,
            logger=                         None,
            loglevel=                       20,
    ) -> None:
        """ copies saved ParaSave POINT from one folder to another """

        if not save_topdir_src: save_topdir_src = cls.SAVE_TOPDIR
        if not save_fn_pfx: save_fn_pfx = cls.SAVE_FN_PFX

        ps_src = cls(
            name=           name_src,
            save_topdir=    save_topdir_src,
            save_fn_pfx=    save_fn_pfx,
            assert_saved=   True,
            logger=         logger,
            loglevel=       loglevel)

        ps_src.name = name_trg
        if save_topdir_trg:
            ps_src.save_topdir = save_topdir_trg

        ps_src.save_point()

    @staticmethod
    def _gxable_check(
            parentA: "ParaSave",
            parentB: Optional["ParaSave"],
    ) -> bool:

        not_gxable_parents = []
        if not parentA['gxable']:
            not_gxable_parents.append(parentA)
        if parentB and not parentB['gxable']:
            not_gxable_parents.append(parentB)

        if not_gxable_parents:
            return False
        return True

    @classmethod
    def gx_saved_point(
            cls,
            name_parentA: str,
            name_parentB: Optional[str],                # if not given makes GX only with parent A
            name_child: str,
            save_topdir_parentA: Optional[str]= None,   # ParaSave top directory of parent A
            save_topdir_parentB: Optional[str]= None,   # ParaSave top directory of parent B
            save_topdir_child: Optional[str]=   None,   # ParaSave top directory of child
            save_fn_pfx: Optional[str]=         None,   # ParaSave POINT file prefix
            logger=                             None,
            loglevel=                           20,
    ) -> None:
        """ performs GX on saved ParaSave POINT """

        if not save_topdir_parentA: save_topdir_parentA = cls.SAVE_TOPDIR
        if not save_topdir_parentB: save_topdir_parentB = save_topdir_parentA
        if not save_topdir_child: save_topdir_child = save_topdir_parentA
        if not save_fn_pfx: save_fn_pfx = cls.SAVE_FN_PFX

        parentA = cls(
            name=           name_parentA,
            save_topdir=    save_topdir_parentA,
            save_fn_pfx=    save_fn_pfx,
            assert_saved=   True,
            logger=         logger,
            loglevel=       loglevel)

        parentB = cls(
            name=           name_parentB,
            save_topdir=    save_topdir_parentB,
            save_fn_pfx=    save_fn_pfx,
            assert_saved=   True,
            logger=         logger,
            loglevel=       loglevel) if name_parentB else None

        if not cls._gxable_check(parentA, parentB):
            raise ParaSaveException('not gxable parents, cannot GX!')

        # make child and save
        point_child = cls.gx_point(
            parentA=    parentA,
            parentB=    parentB,
            name_child= name_child)
        child = parentA
        child.update(point_child)
        child.save_topdir = save_topdir_child
        child.save_point()

    @staticmethod
    def gx_point(
            parentA: "ParaSave",
            parentB: Optional["ParaSave"]=  None,
            name_child: Optional[str]=      None,
            prob_mix=                       0.5,
            prob_noise=                     0.3,
            noise_scale=                    0.1,
            prob_axis=                      0.1,
            prob_diff_axis=                 0.3,
    ) -> POINT:
        """ adds gxable check """

        if not ParaSave._gxable_check(parentA, parentB):
            raise ParaSaveException('not gxable parents, cannot GX')

        child_point = ParaGX.gx_point(
            parentA=        parentA,
            parentB=        parentB,
            name_child=     name_child,
            prob_mix=       prob_mix,
            prob_noise=     prob_noise,
            noise_scale=    noise_scale,
            prob_axis=      prob_axis,
            prob_diff_axis= prob_diff_axis)

        child_point['parents'] = [parentA.parents if parentA.parents else parentA.name]
        if parentB:
            child_point['parents'].append(parentB.parents if parentB.parents else parentB.name)

        return child_point

    @property
    def logger(self):
        return self.__log

    @staticmethod
    def dict_2str(d:dict) -> str:
        """ returns nice string of given dict (mostly for .txt preview save) """
        if d:
            s = ''
            max_len_sk = max([len(k) for k in d.keys()])
            for k, v in sorted(d.items()): s += f'{str(k):{max_len_sk}s} : {str(v)}\n'
            return s[:-1]
        return '--empty point--'

    def __str__(self):
        s = f'{self.__class__.__name__} (ParaSave)\n'
        s += self.dict_2str(self.get_point())
        return s



import random
import string
import time
from typing import List, Any, Optional

from pypaq.exception import PyPaqException
from pypaq.pytypes import NUM
from pypaq.lipytools.moving_average import MovAvg

LINE_UP = '\033[1A'
LINE_CLEAR = '\x1b[2K'



def short_scin(fl:NUM, precision:int=1) -> str:
    """ short (compressed) scientific notation for numbers """
    sh = f'{fl:.{precision}e}'
    sh = sh.replace('+0','')
    sh = sh.replace('+','')
    sh = sh.replace('-0','-')
    return sh


def float_to_str(
        num: float,
        width: int=     7,
        fill: bool=     True,
) -> str:
    """ returns nice string from float, always of given width """

    if width < 5: width = 5
    scientific_decimals = width-6 if width>6 else 0

    fstr = f'{num:.{scientific_decimals}E}'
    if 1000 > num > 0.0001 or num == 0.0:
        fstr = str(num)[:width]

    if fill and len(fstr)<width:
        fstr += ' '*(width-len(fstr))

    return fstr


def stamp(
        year: bool=             False,
        month: bool=            True,
        day: bool=              True,
        hour: bool=             True,
        minutes: bool=          True,
        letters: Optional[int]= 3,
        separator: str=         '_',
) -> str:
    """ prepares timestamp string """

    random.seed(time.time())

    time_format = ''
    if year:            time_format += '%y'
    if month:           time_format += '%m'
    if day:             time_format += '%d'
    if time_format and (hour or minutes): time_format += separator
    if hour:            time_format += '%H'
    if minutes:         time_format += '%M'

    stp = ''
    if time_format:
        stp = time.strftime(time_format)

    if letters:
        if stp: stp += separator
        stp += ''.join([random.choice(string.ascii_letters) for _ in range(letters)])

    return stp


def list_str(ls: List[Any], limit:Optional[int]=200) -> str:
    """ nice string of given list """
    lstr = [str(e) for e in ls]
    lstr = '; '.join(lstr)
    if limit: lstr = lstr[:limit]
    return lstr


def print_nested_dict(
        d: dict,
        ind_scale: int=     2,
        line_limit: int=    100,
) -> None:
    """ prints nice string of nested dict """

    types = {
        dict:   'D',
        list:   'L',
        tuple:  'T',
        str:    'S'}

    def __prn_root(root:dict, ind:int):

        spacer = ' ' * ind * ind_scale

        for key in sorted(list(root.keys())):

            value = root[key]
            tp = types.get(type(value),'')
            value_len_str = str(len(value)) if tp in types.values() else ''

            line = ''
            if tp != 'D':
                line = repr(value)
                if line_limit and len(line)>line_limit:
                    line = f'{line[:line_limit]} ..'
                line = f' : {line}'

            type_len_nfo = f' [{tp}.{value_len_str}]' if tp else ''
            print(f'{spacer}{key}{type_len_nfo}{line}')

            if type(value) is dict:
                __prn_root(root=root[key], ind=ind+1)

    __prn_root(root=d, ind=0)


def printover(sth, clear:int=10) -> None:
    """ prints line over line """
    cls = '' + ' ' * clear
    print(f'\r{sth}{cls}', end='', flush=True)


def printover_terminal(sth) -> None:
    """ prints line over line (terminal alternative) """
    print(sth)
    print(LINE_UP, end=LINE_CLEAR)


class ProgBar:
    """ ProgBar - terminal progress bar,
    speed_avg is an average speed smoothened with moving average,
    speed_cur is a current speed smoothened with moving average,
    guess speed - expected speed value (initial guess), helps to set proper params for the update
    refresh_delay - keeps min delay between refreshes """

    def __init__(
            self,
            total: NUM,
            length: int=            20,
            fill: str=              '█',
            show_fract: bool=       True,
            show_speed_avg: bool=   True,
            show_speed_cur: bool=   True,
            show_eta: bool=         True,
            guess_speed: float=     10.0,
            refresh_delay: float=   1,
    ):
        self.total = total
        self.length = length
        self.fill = fill
        self.show_fract = show_fract
        self.show_speed_avg = show_speed_avg
        self.show_speed_cur = show_speed_cur and self.show_speed_avg
        self.show_eta = show_eta
        self.min_delay_sec = refresh_delay

        f = min(0.5,max(0.01, 1/guess_speed))
        self.speed_a = MovAvg(factor=f) # tot mavg
        self.speed_c = MovAvg(factor=f) # current mavg
        self.n_prev = 0
        self.inc_cached = 0
        self.start_time = time.time()
        self.time_prev = self.start_time

    @staticmethod
    def _speed_to_str(s) -> str:
        if s > 1:
            if s > 100: return f'{int(s)}/s'
            else:       return f'{s:.1f}/s'
        else:           return f'{s:.3f}/s'

    @staticmethod
    def _time_to_str(t) -> str:
        if t > 4000:    return f'{t / 60 / 60:.1f}h'
        else:
            if t > 100: return f'{t / 60:.1f}m'
            else:       return f'{t:.1f}s'

    def __call__(self, n:NUM, prefix:str='', suffix:str=''):

        if n < self.n_prev:
            raise PyPaqException('ProgBar cannot step back with progress (n < prev)')

        if n > self.n_prev:

            time_current = time.time()
            time_passed_prev = time_current - self.time_prev
            if time_passed_prev >= self.min_delay_sec or n >= self.total:

                time_passed_tot = time_current - self.start_time
                self.time_prev = time_current

                progress_factor = n / self.total
                if progress_factor > 1:
                    progress_factor = 1

                filled_length = int(self.length * progress_factor)
                bar_str = self.fill * filled_length + '-' * (self.length - filled_length)

                fract_str = f'{n}/{self.total}' if self.show_fract else ''

                speed_str = ''
                self.speed_a.upd(n / time_passed_tot)
                self.speed_c.upd((n-self.n_prev) / time_passed_prev)
                speed_a = self.speed_a()
                if self.show_speed_avg:
                    speed_str = self._speed_to_str(speed_a)
                if self.show_speed_cur:
                    speed_str += f'[{self._speed_to_str(self.speed_c())}]'

                eta_str = ''
                if self.show_eta:
                    if speed_a > 0:
                        eta = (self.total - n) / speed_a
                        if eta < 0:
                            eta = 0
                        eta_str = self._time_to_str(eta)
                    else:                 eta_str = '---'
                    eta_str = f'ETA:{eta_str}'

                detailsL = [fract_str,speed_str,eta_str]
                detailsL = [e for e in detailsL if e]
                details_str = f'{" ".join(detailsL)} ' if detailsL else ''

                elapsed_str = ''
                if progress_factor == 1 and self.show_eta:
                    elapsed_str = '-- TOT:' + self._time_to_str(time_passed_tot)
                    elapsed_str += f' {self._speed_to_str(self.total / time_passed_tot)}'


                printover(f'{prefix}|{bar_str}|{progress_factor * 100:.1f}% '
                          f'{details_str}{elapsed_str}{suffix}')

                if progress_factor == 1:
                    print()

                self.n_prev = n
                self.inc_cached = 0

    def inc(self, prefix:str='', suffix:str=''):
        """ increase by 1 """
        self.inc_cached += 1
        self(n=self.n_prev + self.inc_cached, prefix=prefix, suffix=suffix)



from abc import ABC, abstractmethod
from multiprocessing import cpu_count, Process, Queue, Value
import psutil
from queue import Empty
import time
from typing import Any, Optional, Union

from pypaq.exception import PyPaqException
from pypaq.lipytools.pylogger import get_pylogger


class MPythonException(PyPaqException):
    pass


class QMessage:
    """ QMessage is a type of object (message) sent between processes with Ques """
    def __init__(self, type:str, data:Optional[Any]=None):
        self.type = type
        self.data = data

    def __str__(self):
        return f'{self.__class__.__name__}, type:{self.type}, data:{self.data}'

# https://github.com/vterron/lemon/commit/9ca6b4b1212228dbd4f69b88aaf88b12952d7d6f
class SharedCounter:

    def __init__(self, n:int=0):
        self.count = Value('i', n)

    def increment(self, n:int=1):
        with self.count.get_lock():
            self.count.value += n

    @property
    def value(self) -> int:
        return self.count.value


class Que:
    """ MultiProcessing Queue that:
    - manages its size
    - accepts messages of QMessage type """

    def __init__(self):
        self.q = Queue()
        self.size = SharedCounter(0)

    def put(self, msg:QMessage, **kwargs):
        if not isinstance(msg, QMessage):
            raise MPythonException(f'\'msg\' should be type of QMessage, but is {type(msg)}')
        self.size.increment(1)
        self.q.put(msg, **kwargs)

    def get(self, block:bool=True, timeout:Optional[float]=None) -> Optional[QMessage]:
        try:
            msg = self.q.get(block=block, timeout=timeout)
            self.size.increment(-1)
        except Empty:
            return None
        if not isinstance(msg, QMessage):
            raise MPythonException(f'\'msg\' should be type of QMessage, but is {type(msg)}')
        return msg

    @property
    def empty(self) -> bool:
        return not self.qsize

    @property
    def qsize(self) -> int:
        return self.size.value


class ExProcess(Process, ABC):
    """ Exception Managed Process
    implements basic exceptions management """

    def __init__(
            self,
            ique: Optional[Que]=            None,
            oque: Optional[Que]=            None,
            name: Optional[Union[str,int]]= None,
            raise_KeyboardInterrupt=        False,
            raise_Exception=                False,
            logger=                         None,
            loglevel=                       30):
        """
        :param ique:
            input Que, not used by this class, but may be used by child Classes
        :param oque:
            output Que, sends messages from the ExProcess,
            when Exception or KeyboardInterrupt occurs message is sent here
        :param name:
            identifies ExProcess, for None is taken from Process.name
        :param raise_KeyboardInterrupt:
            results in immediate exit (without running post exception handler)
        :param raise_KeyboardInterrupt:
            raises Exception output (debug) """

        super().__init__(target=self.__run)

        if name is not None:
            self.name = str(name)

        if not logger:
            logger = get_pylogger(
                name=       self.name,
                folder=     None,
                level=      loglevel)
        self.logger = logger

        self.ique = ique
        self.oque = oque
        self.raise_KeyboardInterrupt = raise_KeyboardInterrupt
        self.raise_Exception = raise_Exception

        self.logger.info(f'*** {self.name} (ExProcess) *** initialized')

    def __run(self):
        """ process target method,
        wraps exprocess_method() with try / except = exception handling """
        try:
            self.logger.debug(f'> ExProcess ({self.name}, pid:{self.pid}) - started exprocess_method()')
            self.exprocess_method()
            self.logger.debug(f'> ExProcess ({self.name}, pid:{self.pid}) - finished exprocess_method()')

        except KeyboardInterrupt:

            # allows intermediate interrupt
            if self.raise_KeyboardInterrupt:
                raise KeyboardInterrupt

            self.__exception_handle('(Base)KeyboardInterrupt')
            if self.raise_Exception:
                raise KeyboardInterrupt

        except Exception as e:

            self.__exception_handle(str(e))
            if self.raise_Exception:
                raise e

    @abstractmethod
    def exprocess_method(self):
        """ method run in a process, to be implemented """
        pass

    def __exception_handle(self, e_name:str):
        """ when exception occurs, message with exception data is put on the output que """
        if self.oque is not None:
            self.oque.put(QMessage(
                type=   f'Exception:{e_name}, ExProcess {self.name} (pid:{self.pid})',
                data=   self.name)) # returns name here to allow process identification
        self.logger.warning(f'> ExProcess ({self.name}) halted by Exception:{e_name}')
        self.after_exception_handle_run()

    def after_exception_handle_run(self):
        """ this method may be implemented and will be run after inside exception handler """
        pass

    @property
    def alive(self):
        """ not spawned process is not alive and cannot be joined,
        when proces alive: True->False it may be join()-ed """
        if self.closed:
            return False
        return self.is_alive()

    def terminate(self):
        if self.alive:
            super().terminate()
        while self.alive:
            time.sleep(0.01)

    def kill(self):
        if self.alive:
            super().kill()
        while self.alive:
            time.sleep(0.01)

    @property
    def closed(self):
        return self._closed

    def __str__(self):
        pid = self.pid if not self.closed else '<-closed->'
        exitcode = self.exitcode if not self.closed else '<-closed->'
        mem_nfo = '---'
        if self.alive:
            mem_mb = int(psutil.Process(self.pid).memory_info().rss / 1024 ** 2)
            mem_nfo = f'{mem_mb}MB'
        return (f'{super().__str__()}\n'
                f'> pid:            {pid}\n'
                f'> parent(pid):    {self._parent_pid}\n'
                f'> alive:          {self.alive}\n'
                f'> closed:         {self.closed}\n'
                f'> exitcode:       {exitcode}\n'
                f'> mem:            {mem_nfo}')


def sys_res_nfo():
    vm = psutil.virtual_memory()
    gb = 1024 ** 3
    return {
        'cpu_count':    cpu_count(),
        'cpu_used_%':   psutil.cpu_percent(interval=5), # over last 5 sec
        'mem_total_GB': vm.total / gb,
        'mem_used_%':   vm.percent}


import time
from typing import Dict, List, Tuple, Optional

from pypaq.exception import PyPaqException


class TimeRep:
    """ time reporting helper """

    def __init__(self):
        self.start_now()

    def start_now(self):
        """ resets and starts """
        self.tr: Dict[str, Tuple[float, Optional[TimeRep]]] = {}
        self.stime = time.time()
        self.stime_start = self.stime
        self.sub_tr_keys = []

    def log(self, interval_name:str, interval_tr:Optional["TimeRep"]=None):
        ct = time.time()
        if interval_name in self.tr:
            raise PyPaqException(f'interval_name {interval_name} already reported')
        if interval_tr:
            irep = interval_tr.get_report(sub=True)
            for k in irep:
                if k in self.sub_tr_keys:
                    raise PyPaqException(f'sub interval_name {k} already reported')
                self.sub_tr_keys.append(k)
        self.tr[interval_name] = (ct - self.stime, interval_tr)
        self.stime = ct

    def get_report(self, sub=False, total=False) -> Dict[str,float]:
        rep = {}
        for k in self.tr:
            rep[k] = self.tr[k][0]
            if sub and self.tr[k][1]:
                sub_rep = self.tr[k][1].get_report(sub=True)
                for sk in sub_rep:
                    rep[f'---{sk}'] = sub_rep[sk]
        if total:
            rep['___total time'] = time.time() - self.stime_start
        return rep

    @staticmethod
    def _get_line(n,t):
        return f'{n:30}: {t:9.3f}s'

    def _get_report_lines(self) -> List[str]:
        rl = []
        for n,(t,itr) in self.tr.items():
            rl.append(self._get_line(n,t))
            if itr:
                rl += [f'---{l}' for l in itr._get_report_lines()]
        return rl

    def __str__(self):
        s = "\n".join(self._get_report_lines())
        return s + f"\n{self._get_line('___total time', time.time() - self.stime_start)}"


from pypaq.exception import PyPaqException


# double hinge function _/**
def double_hinge(
        s_val: float,
        e_val: float,
        a: int,
        b: int,
        counter: int,
) -> float:
    """ returns:
    - s_val for counter <= a
    - linear interpolation from s_val to e_val in range (a;b)
    - e_val for counter >= b """

    if b < a or b == a and s_val != e_val:
        raise PyPaqException('bad arguments!')

    if counter <= a:
        return s_val
    if counter >= b:
        return e_val
    y = (counter - a) / (b - a)
    return s_val + (e_val-s_val) * y


import math
import numpy as np
from scipy import stats
from typing import Tuple, Union, Iterable


class ValuesArray:
    """ ValuesArray is a np.ndarray
    which behaves like a python List
    (only some interfaces are implemented).
    It is optimized for constant grow and large sizes. 
    Additionally, supports:
    - mean and h95 calculation """

    def __init__(
            self,
            init_size: int= 100000,
            dtype: type=    np.float32,
    ):
        self.dtype = dtype
        self._val = np.zeros(shape=init_size, dtype=self.dtype)
        self.size = 0

    def _grow(self):
        cs = self._val.shape[0]
        _wr = self._val
        self._val = np.zeros(shape=cs*2, dtype=self.dtype)
        self._val[:cs] = _wr

    # *** List interface ***************************************************************************

    def __len__(self) -> int:
        return self.size

    def __iadd__(self, other:Union[Iterable,"ValuesArray"]):

        if isinstance(other, ValuesArray):
            other = other.get_array()
        _other_arr = np.asarray(other, dtype=self.dtype)

        while len(self._val) - self.size < len(_other_arr):
            self._grow()

        self._val[self.size:self.size+len(_other_arr)] = _other_arr
        self.size += len(_other_arr)
        return self

    def __str__(self):
        return str(self.get_array())

    def append(self, v):
        self._val[self.size] = v
        self.size += 1
        if self.size == self._val.shape[0]:
            self._grow()

    # *** additional interface *********************************************************************

    def get_array(self) -> np.ndarray:
        return self._val[:self.size]

    def mean_h95(self) -> Tuple[float,float]:
        mean = float(self._val[:self.size].mean())
        std = self._val[:self.size].std()
        sem = std / math.sqrt(self.size)
        h95 = float(sem * stats.t.ppf(0.975, self.size - 1))
        return mean, h95


import math
import numpy as np
import pandas as pd
from scipy import stats
from typing import Dict, Tuple

from pypaq.pytypes import NPL


# min, avg, max
def mam(vals:NPL) -> Tuple[float,float,float]:
    if len(vals): return min(vals), sum(vals) / len(vals), max(vals)
    else:         return 0.0, 0.0, 0.0

# mean, median, std, SEM, h95, min, max
def msmx(vals:NPL) -> Dict:
    arr = np.asarray(vals) if type(vals) is list else vals
    _mean, _median, _std, _min, _max = float(np.mean(arr)), float(np.median(arr)), float(np.std(arr)), float(np.min(arr)), float(np.max(arr))
    _sem = _std / math.sqrt(len(vals))
    _h95 = _sem * stats.t.ppf(0.975, len(vals) - 1) # 0.975 is 1 + 0.95 / 2
    ret_dict = {'mean':_mean, 'median':_median, 'std':_std, 'sem':_sem, 'h95':_h95, 'min':_min, 'max':_max}
    ret_dict['string'] = ' '.join([f'{k}:{ret_dict[k]:.5f}' for k in ret_dict])
    return ret_dict

# some stats with pandas
def stats_pd(
        val_list :NPL,
        n_percentiles=  10) -> str:
    s = f'{pd.Series(val_list).describe(percentiles=[0.1*n for n in range(1,n_percentiles)])}'
    return s[:s.rfind('\n')]


import csv
import gzip
import json
import os
import pickle
import shutil
import yaml
from pathlib import Path
import sys
from typing import Union, Dict, List, Optional

from pypaq.exception import PyPaqException


def r_text(file_path:str, raise_exception=False) -> Optional[str]:
    if not os.path.isfile(file_path):
        if raise_exception:
            raise FileNotFoundError(f'file {file_path} not exists!')
        return None
    with open(file_path, 'r', encoding='utf-8') as file:
        return file.read()


def w_text(text:str, file_path:str):
    with open(file_path, 'w') as file:
        return file.write(text)


def r_pickle(file_path, obj_type=None, raise_exception=False):
    """ reads pickle
    if obj_type is given checks for compatibility with given type """
    if not os.path.isfile(file_path):
        if raise_exception:
            raise FileNotFoundError(f'file {file_path} not exists!')
        return None

    compressed = False
    with open(file_path, 'rb') as f:
        if f.read(2) == b'\x1f\x8b': # gzip magic number
            compressed = True

    op_fn = gzip.open if compressed else open
    with op_fn(file_path, 'rb') as file:
        obj = pickle.load(file)

    if obj_type:
        if not type(obj) is obj_type:
            raise PyPaqException(f'ERROR: obj from file is not {str(obj_type)} type')
    return obj


def w_pickle(obj, file_path, compressed=False):
    op_fn = gzip.open if compressed else open
    with op_fn(file_path, 'wb') as file:
        pickle.dump(obj, file)


def r_json(file_path, raise_exception=False) -> Optional[Union[Dict,List]]:
    if not os.path.isfile(file_path):
        if raise_exception:
            raise FileNotFoundError(f'file {file_path} not exists!')
        return None
    with open(file_path, 'r', encoding='utf-8') as file:
        return json.load(file)


def w_json(data:Union[Dict,List], file_path):
    with open(file_path, 'w', encoding='utf-8') as file:
        json.dump(data, file, indent=4, ensure_ascii=False)


def r_jsonl(file_path, raise_exception=False) -> Optional[List]:
    if not os.path.isfile(file_path):
        if raise_exception:
            raise FileNotFoundError(f'file {file_path} not exists!')
        return None
    with open(file_path, 'r', encoding='utf-8') as file:
        return [json.loads(line) for line in file]


def w_jsonl(data:List, file_path):
    with open(file_path, 'w', encoding='utf-8') as file:
        for d in data:
            json.dump(d, file, ensure_ascii=False)
            file.write('\n')


def r_csv(file_path, raise_exception=False) -> Optional[List]:
    if not os.path.isfile(file_path):
        if raise_exception:
            raise FileNotFoundError(f'file {file_path} not exists!')
        return None
    csv.field_size_limit(sys.maxsize)
    with open(file_path, newline='') as f:
        reader = csv.reader(f)
        return [row for row in reader]


def w_csv(data:List[List], file_path):
    with open(file_path, 'w', newline='') as csvfile:
        writer = csv.writer(csvfile)
        writer.writerows(data)


def r_yaml(file_path, raise_exception=False):
    if not os.path.isfile(file_path):
        if raise_exception:
            raise FileNotFoundError(f'file {file_path} not exists!')
        return None
    with open(file_path) as file:
        return yaml.load(file, yaml.Loader)


def get_dir(path: Union[str, Path]):
    path = str(path)
    path_split = path.split('/')
    if path_split[-1].find('.') != -1: path_split = path_split[:-1]
    return '/'.join(path_split)


def prep_folder(path:Union[str,Path], flush_non_empty=False):
    """ prepares folder
    path may be a file path -> folder path is extracted """
    folder_path = get_dir(path)
    if folder_path: # in case folder_path == ''
        if flush_non_empty and os.path.isdir(folder_path):
            shutil.rmtree(folder_path)
        os.makedirs(folder_path, exist_ok=True)


def list_dir(path: Union[str, Path]) -> Dict[str,List]:
    ls = os.listdir(path)
    lsD = {'files':[], 'dirs':[]}
    for e in ls:
        lsD['files' if os.path.isfile(f'{path}/{e}') else 'dirs'].append(e)
    return lsD


def get_requirements(file_path:str='requirements.txt') -> List[str]:
    file_text = r_text(file_path, raise_exception=True)
    file_lines = file_text.split('\n')
    return [l.strip() for l in file_lines]




import inspect
from typing import Any, Dict, List, Tuple, Callable, Optional, Union, Type

from pypaq.exception import PyPaqException
from pypaq.lipytools.printout import float_to_str

AXIS =  str                     # axis type (parameter name)
P_VAL = Union[float, int, Any]  # point value (parameter value)
POINT = Dict[AXIS, P_VAL]       # POINT ia a dict {parameter: value}


""" 
    PSDD - Parameters Space Definition Dict
    dictionary that defines space of POINTS {axis(parameter name): list or tuple or value}
        - list of ints or floats defines continuous range
        - tuple may contain elements of any type (even non-numeric), may have one
        - any other type is considered to be a constant (single value)

        example:
        {   'a':    [0.0, 1],               # range of floats
            'b':    (-1,-7,10,15.5,90,30),  # set of num(float) values, num will be sorted
            'c':    ('tat','mam','kot'),    # set of diff values
            'd':    [0,10],                 # range of ints
            'e':    (-2.0,2,None)}          # set of diff values
            'f':    (16.2,)}                # single value
"""

RANGE = List[P_VAL] or Tuple[P_VAL] or P_VAL    # axis range type (range of parameter)
PSDD  = Dict[AXIS, RANGE]                       # Parameters Space Definition Dict {parameter: range}}


class PMSException(PyPaqException):
    pass


def point_str(p:POINT) -> str:
    """ prepares nice string of POINT """

    avL = []
    for axis in sorted(list(p.keys())):

        val = p[axis]

        val_str = ''
        if type(val) is float:
            val_str = float_to_str(val)
        if type(val) is bool:
            val_str = f'{str(val):5}'
        if not val_str:
            val_str = str(val)

        avL.append(f'{axis}:{val_str}')

    return '{' + ' '.join(avL) + '}'


def get_params(function:Callable) -> Dict:
    """ prepares function parameters dictionary """

    params_dict = {
        'without_defaults': [],
        'with_defaults':    {}}

    if function:

        specs = inspect.getfullargspec(function)
        #print(specs)

        params = specs.args + specs.kwonlyargs

        vals = []
        if specs.defaults:
            vals += list(specs.defaults)
        if specs.kwonlydefaults:
            vals += list(specs.kwonlydefaults.values())

        while len(params) > len(vals):
            params_dict['without_defaults'].append(params.pop(0))

        params_dict['with_defaults'] = {k: v for k,v in zip(params,vals)}

    return params_dict


def get_class_init_params(cl:Type) -> Dict:
    """ prepares class.__init__ parameters dictionary
    recurrently goes down to base classes """

    def _update_params_dict(
            without_defaults: List,
            with_defaults: Dict):
        """ in-place updates current params_dict """

        # cross-remove
        for k in without_defaults:
            if k in params_dict['with_defaults']:
                params_dict['with_defaults'].pop(k)
        for k in with_defaults:
            if k in params_dict['without_defaults']:
                params_dict['without_defaults'].remove(k)

        for p in without_defaults:
            if p not in params_dict['without_defaults']:
                params_dict['without_defaults'].append(p)
        params_dict['with_defaults'].update(with_defaults)

    params_dict = {
        'without_defaults': [],
        'with_defaults':    {}}

    bases = list(cl.__bases__)
    if object in bases:
        bases.remove(object)

    for bcl in bases:
        bcl_pd = get_class_init_params(bcl)
        _update_params_dict(
            without_defaults=   bcl_pd['without_defaults'],
            with_defaults=      bcl_pd['with_defaults'])

    cl_pd = get_params(cl.__init__)
    _update_params_dict(
        without_defaults=   cl_pd['without_defaults'],
        with_defaults=      cl_pd['with_defaults'])

    return params_dict


def point_trim(
        fc: Optional[Union[Callable,Type]],
        point: POINT,
        remove_self= True,  # removes self in case of class methods
) -> POINT:
    """ prepares sub-POINT trimmed to function params (given wider POINT) """

    if fc is None:
        return {}

    pms = get_params(fc) if inspect.isfunction(fc) else get_class_init_params(fc)

    valid_keys = pms['without_defaults'] + list(pms['with_defaults'].keys())

    if remove_self and 'self' in valid_keys:
        valid_keys.remove('self')

    func_dna = {k: point[k] for k in point if k in valid_keys} # filter to get only params accepted by func

    return func_dna



def whitespace_normalization(text:str, remove_nlines=True) -> str:
    """ normalizes whitespace characters & newlines -> single space
    - remove_nlines:
        False -> any combination of whitespaces with \n will be replaced with single \n
        True ->  any combination of whitespaces with \n will be replaced with ' ' """

    def _line_whitespace_normalization(txt: str) -> str:
        if txt: txt = ' '.join(txt.split())
        return txt

    lines = [text] if remove_nlines else text.split('\n')
    lines = [_line_whitespace_normalization(txt=line) for line in lines]
    lines = [l for l in lines if l]  # remove empty lines
    return '\n'.join(lines)


import logging
from typing import Optional, Union, Tuple

from pypaq.lipytools.printout import stamp
from pypaq.lipytools.files import prep_folder


class PyLogger(logging.Logger):
    pass

logging.setLoggerClass(PyLogger)


def get_pylogger(
        name: Optional[str]=                None,
        add_stamp=                          True,
        folder: Optional[str]=              None,
        level=                              logging.INFO,
        flat_child: bool=                   False,
        format: Union[Tuple[str,str],str]=  '%(asctime)s {%(filename)20s:%(lineno)4d} p%(process)s %(levelname)s: %(message)s',
        to_stdout=                          True,
) -> logging.Logger:
    """ returns formatted logging.Logger
    - add_stamp:    prevents merging loggers of same name
    - folder:       writes logfile to folder if given
    - flat_child:   forces child of this logger created with get_child() to be same level
    - format:       may be given as a str or Tuple[str,str] (fmt,datefmt) """

    if not name and folder:
        name = 'pylogger'
    if name and add_stamp:
        name += '_' + stamp()

    if type(format) is not tuple:
        format = (format,)

    formatter = logging.Formatter(*format)

    logger = logging.getLogger(name)
    logger.setLevel(level)

    ### manage handlers

    # if logger has handlers new handlers of existing type won't be added
    for h in logger.handlers:
        if type(h) is logging.FileHandler:
            folder = None
        if type(h) is logging.StreamHandler:
            to_stdout = False

    if folder:
        prep_folder(folder)
        fh = logging.FileHandler(f'{folder}/{name}.log')
        fh.setFormatter(formatter)
        logger.addHandler(fh)

    if to_stdout:
        sh = logging.StreamHandler()
        sh.setFormatter(formatter)
        logger.addHandler(sh)

    logger.flat_child = flat_child

    return logger


def get_child(logger, name:Optional[str], change_level:int=0) -> logging.Logger:
    """ returns child with optionally changed level """

    clogger = logger.getChild(name or '_child')
    clogger.flat_child = logger.flat_child

    if change_level != 0 and not logger.flat_child:
        lvl = clogger.getEffectiveLevel()
        lvl_child = min(lvl + change_level, logging.WARNING)
        clogger.setLevel(lvl_child)

    return clogger


import numpy as np
from pypaq.exception import PyPaqException
from pypaq.pytypes import NUM
from typing import Optional


class MovAvg:
    """ moving average,
    updates self.value with factor,
    optionally averages first 1/factor/10 values, which improves early estimation and diminishes bias """

    def __init__(
            self,
            factor: NUM=                0.1,    # (0.0;1.0>
            first_avg: bool=            True,   # to average first 1/factor values
            init_value: Optional[NUM]=  None,   # set starting value
            init_weight: int=           10,     # set weight (count) for starting value
    ):
        self.value: Optional[NUM] = init_value

        if not 0 < factor <= 1:
            raise PyPaqException('factor should: 0 < factor <= 1')

        self.factor = factor
        self.upd_ix = 0 if self.value is None else init_weight

        self.first_avg = first_avg
        self.n_first = int(1/self.factor/10)        # how many first values will be taken into average
        self.first_np = np.zeros(self.n_first)      # cache first here
        self.first_np[:self.upd_ix] = self.value

    def upd(self, val:NUM):

        if self.first_avg and self.upd_ix < self.n_first:
            self.first_np[self.upd_ix] = val
            self.value = float(np.mean(self.first_np[:self.upd_ix + 1]))
        else:
            if self.value is None:
                self.value = val
            else:
                self.value = (1-self.factor)*self.value + self.factor*val

        self.upd_ix += 1

        return self.value

    def __call__(self) -> NUM:
        if self.value is None:
            raise PyPaqException('MovAvg not updated yet, value unknown')
        return self.value